from typing import Text

from altimate_models.base.extractor import SQLAlchemyExtractor
from altimate_models.base.metadata_profiler import SqlAlchemyMetadataProfiler
from altimate_models.base.source import DataSource
from altimate_models.postgresql.dialect import PostgreSQLSourceDialect
from altimate_models.postgresql.extractor import PostgreSQLExtractor
from altimate_models.postgresql.profiler import PostgreSQLProfiler
from altimate_models.s3.source import S3Source

from altimate_models.duckdb.extractor import DuckDBExtractor
from altimate_models.s3.csv_source_dialect import CsvSourceDialect
from altimate_models.s3.json_source_dialect import JsonSourceDialect
from altimate_models.s3.parquet_source_dialect import ParquetSourceDialect
from altimate_models.duckdb.metadata_profiler import DuckDBMetadataProfiler
from altimate_models.s3.profiler import S3Profiler
from altimate_models.tableau.extractor import TableauExtractor
from altimate_profiler.transformer.duckdb_schema import DuckDBSchemaTransformer
from altimate_models.shared_models import Policy
from altimate_models.snowflake.dialect import SnowflakeSourceDialect
from altimate_models.snowflake.profiler import SnowflakeProfiler

from altimate_profiler.exceptions import AltimateDataStoreNotSupported
from altimate_profiler.transformer.base import Transformer
from altimate_profiler.transformer.schema import SchemaTransformer


class SchemaTransformerFactory:
    SOURCE_MAP = {
        "snowflake": SchemaTransformer,
        "postgres": SchemaTransformer,
        "s3": DuckDBSchemaTransformer,
    }
    TABLE_SOURCES = ["snowflake", "postgres"]
    FILE_SOURCES = ["s3"]

    @classmethod
    def create(cls, data_source_type: Text) -> Transformer:
        transformer = cls.SOURCE_MAP.get(data_source_type)
        if transformer is None:
            raise AltimateDataStoreNotSupported("Invalid data store type")
        return transformer


class MetadataFactory:
    DATA_SOURCE_TO_PROFILER = {
        "snowflake": SqlAlchemyMetadataProfiler,
        "postgres": SqlAlchemyMetadataProfiler,
        "s3": DuckDBMetadataProfiler,
    }

    @classmethod
    def create(
        cls,
        data_source: DataSource,
        policy: Policy,
        resource_name: Text,
    ):
        profiler = cls.DATA_SOURCE_TO_PROFILER.get(data_source.type)
        dialect = DialectFactory.create(data_source)
        if not profiler:
            raise AltimateDataStoreNotSupported("Data source is not supported!")

        return profiler(data_source, dialect, policy, resource_name)


class ProfilerFactory:
    DATA_SOURCE_TO_PROFILER = {
        "snowflake": SnowflakeProfiler,
        "postgres": PostgreSQLProfiler,
        "s3": S3Profiler,
    }

    @classmethod
    def create(cls, data_source: DataSource, policy: Policy, use_account_usage: bool = False):
        profiler = cls.DATA_SOURCE_TO_PROFILER.get(data_source.type)
        if not profiler:
            raise AltimateDataStoreNotSupported("Data source is not supported!")

        return profiler(data_source, policy, use_account_usage)


class DialectFactory:
    DATA_SOURCE_TO_DIALECT = {
        "snowflake": SnowflakeSourceDialect,
        "postgres": PostgreSQLSourceDialect,
        "s3": {
            "csv": CsvSourceDialect,
            "parquet": ParquetSourceDialect,
            "json": JsonSourceDialect,
        }
    }

    @classmethod
    def create(cls, data_source: DataSource):
        if isinstance(
            data_source, S3Source
        ):  # use stronger match than string comparison `data_source.type == "s3":`
            dialect = cls.DATA_SOURCE_TO_DIALECT.get('s3').get(
                data_source.file_format
            )
        else:
            dialect = cls.DATA_SOURCE_TO_DIALECT.get(data_source.type)

        if not dialect:
            raise AltimateDataStoreNotSupported("Data source is not supported!")
        return dialect


class ExtractorFactory:
    DATA_SOURCE_TO_Extractor = {
        "snowflake": SQLAlchemyExtractor,
        "postgres": PostgreSQLExtractor,
        "s3": DuckDBExtractor,
        "tableau": TableauExtractor,
    }

    @classmethod
    def create(cls, data_source: DataSource):
        extractor = cls.DATA_SOURCE_TO_Extractor.get(data_source.type)
        if not extractor:
            raise AltimateDataStoreNotSupported("Data source is not supported!")
        return extractor



import sys
from typing import Dict, Optional, Text

try:
    import boto3
    import click
except ImportError:
    pass

from altimate_models.base.source import DataSource
from altimate_models.shared_models import Resource, ResourceType, S3Resource
from altimate_profiler.exceptions import AltimateDataStoreConnectionException
from altimate_profiler.utils import green_bold_string, red, sha256_encoding


class S3Source(DataSource):
    NAME = "s3"

    aws_access_key_id: Optional[Text]
    aws_secret_access_key: Optional[Text]
    region: Text
    bucket: Text
    prefix: Text
    file_format: Text
    options: Optional[Dict[Text, Text]]

    s3_endpoint: Optional[Text] = None
    s3_use_ssl: Optional[Text] = None
    s3_url_style: Optional[Text] = None

    def get_connection_string(self, *args, **kwargs) -> Text:
        connection_str = """
INSTALL httpfs;
LOAD httpfs;
SET s3_region='{region}';
SET s3_access_key_id='{access_key}';
SET s3_secret_access_key='{secret_key}';
        """

        if self.s3_endpoint:
            connection_str += f"\nSET s3_endpoint='{self.s3_endpoint}';"
        if self.s3_use_ssl:
            connection_str += f"\nSET s3_use_ssl='{self.s3_use_ssl}';"
        if self.s3_url_style:
            connection_str += f"\nSET s3_url_style='{self.s3_url_style}';"

        return connection_str.format(
            region=self.region,
            access_key=self.aws_access_key_id,
            secret_key=self.aws_secret_access_key,
        )

    @classmethod
    def get_connection_information(cls, data_store_config: Dict):
        data_store_config[
            "connection_info"
        ] = f"{data_store_config['bucket']}.{data_store_config['region']}/{data_store_config['prefix']}?file_format={data_store_config['file_format']}"
        data_store_config["connection_hash"] = sha256_encoding(
            data_store_config["connection_info"]
        )
        return data_store_config

    @classmethod
    def prompt(cls, data_store_config: Dict):
        if "click" not in sys.modules:
            raise ImportError(
                "click is not installed. Please install click to connect to S3."
            )
        data_store_config["bucket"] = click.prompt("bucket", type=str)
        data_store_config["aws_access_key_id"] = click.prompt(
            "aws_access_key_id", type=str
        )
        data_store_config["aws_secret_access_key"] = click.prompt(
            "aws_secret_access_key", type=str, hide_input=True
        )
        data_store_config["region"] = click.prompt("region", type=str, default="")
        data_store_config["prefix"] = click.prompt("Path to file", type=str, default="")
        data_store_config["file_format"] = click.prompt(
            "file format", type=click.Choice(["csv", "parquet", "json"])
        )
        data_store_config["options"] = {}
        if data_store_config["file_format"] == "csv":
            data_store_config["options"]["header"] = click.prompt(
                "header", type=bool, default=True
            )
            data_store_config["options"]["delim"] = (
                '"' + click.prompt("delimiter", type=str, default=",") + '"'
            )

        data_store_config = cls.get_connection_information(data_store_config)
        return S3Source(**data_store_config)

    def cred_prompt(self):
        if "click" not in sys.modules:
            raise ImportError(
                "click is not installed. Please install click to connect to S3."
            )
        self.aws_access_key_id = click.prompt("aws_access_key_id", type=str)
        self.aws_secret_access_key = click.prompt(
            "aws_secret_access_key", type=str, hide_input=True
        )
        conn_info = self.get_connection_information(self.__dict__)
        self.connection_info = conn_info["connection_info"]
        self.connection_hash = conn_info["connection_hash"]

    def test_connection(self):
        if "boto3" in sys.modules:
            session = boto3.Session(
                aws_access_key_id=self.aws_access_key_id,
                aws_secret_access_key=self.aws_secret_access_key,
            )

            # Create a client for S3 service
            s3_client = session.client("s3")

            try:
                # Test the connection by listing objects in the bucket
                response = s3_client.list_objects_v2(Bucket=self.bucket)
                print(green_bold_string("Successfully connected to S3!"))
            except Exception as e:
                print(red("Connection to S3 failed."))
                print("Check your credentials and try again.")
                raise AltimateDataStoreConnectionException(
                    f"Connection to {self.name} failed."
                    f" Please check the credentials"
                )
        else:
            print("boto3 is not installed. Please install boto3 to connect to S3.")

    def drop_credentials(self):
        config_dict = self.dict()
        aws_access_key_id = config_dict.pop("aws_access_key_id")
        aws_secret_access_key = config_dict.pop("aws_secret_access_key")
        return config_dict

    def get_resource(self):
        return Resource(
            resource_type=ResourceType.FILE,
            resource=S3Resource(
                bucket=self.bucket,
                path=self.prefix,
                format=self.file_format,
                options=self.options,
            ),
        )

    @classmethod
    def get_config_from_url(cls, connection_info):
        file_format = ""

        if "?" in connection_info:
            resource_path, params = connection_info.split("?")
            for param in params.split("&"):
                key, value = param.split("=")
                if key == "file_format":
                    file_format = value
        else:
            resource_path = connection_info
        bucket, region = resource_path.split("/")[0].split(".")
        prefix = "/".join(resource_path.split("/")[1:])
        return {
            "bucket": bucket,
            "region": region,
            "prefix": prefix,
            "file_format": file_format,
        }

    def get_identifier_and_fqn(self):
        identifier = (
            f"{self.type}://{self.bucket}/{self.prefix}?file_format={self.file_format}"
        )
        fqn = f"bucket={self.bucket}"
        return identifier, fqn

    @classmethod
    def get_resource_config(cls, resource_config: Dict):
        filters = resource_config.get("filters")
        identifier = resource_config.get("identifier")
        data_store_type, path = identifier.split("://")
        options = resource_config.get("options")
        resource_path, params = path.split("?")
        file_format = params.split("=")[1]
        bucket = resource_path.split("/")[0]
        path = "/".join(resource_path.split("/")[1:])

        resource = S3Resource(
            bucket=bucket,
            path=path,
            format=file_format,
            options=options,
        )

        return Resource(
            resource_type=ResourceType.FILE, resource=resource, filters=filters
        )

    @classmethod
    def map_config_to_source(cls, config: Dict):
        connection_info = config.get("connection_info")
        if not connection_info:
            return config
        s3_config = cls.get_config_from_url(connection_info)
        config["bucket"] = s3_config["bucket"]
        config["region"] = s3_config["region"]
        config["prefix"] = s3_config["prefix"]
        config["file_format"] = config.get("file_format") or s3_config["file_format"]
        config.pop("connection_info")
        return config

    def override_schema(self, schema):
        return schema



from typing import List, Text

from altimate_profiler.metrics.table_metric_type import TableMetricType
from altimate_models.shared_models import Policy
from altimate_profiler.builders.base_builder import Base


class S3TableQueryBuilder(Base):
    def __init__(
            self,
            dialect,
            policy: Policy,
    ):
        self.policy = policy
        self.dialect = dialect


    def _gen_sql(self, metric, params, table):
        attr = getattr(self.dialect, metric._value_)
        return attr(table, params)

    def compile(self) -> List[Text]:
        metrics_queries = []
        all_filters = ' AND '.join(x for x in [self.policy.resource0.filters, self.policy.filters] if x)
        table = self.dialect._table_sql(self.policy.resource0.resource.bucket, self.policy.resource0.resource.path, self.policy.resource0.resource.options, all_filters)
        for term in self.policy.terms:
            metrics_queries.append(self._gen_sql(TableMetricType[term.metric.upper()], term.params, table))
        return metrics_queries



import decimal

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from altimate_models.base.source import DataSource
    
from altimate_profiler.builders.utils import map_type_to_standard_types, infer_type
from altimate_models.duckdb.profiler import DuckDBProfiler
from altimate_models.s3.extractor import S3Extractor
from altimate_profiler.metrics.metadata_metric_type import MetadataMetricType
from altimate_models.shared_models import Policy
from altimate_profiler.builders.metadata_query_builder import MetadataQueryBuilder


class DuckDBMetadataProfiler(DuckDBProfiler):

    COMMON_METRICS = [
        MetadataMetricType.CARDINALITY, MetadataMetricType.NULL_PERCENTAGE
    ]

    NUMERIC_METRICS = [
        MetadataMetricType.NUMERIC_MAX, MetadataMetricType.NUMERIC_MIN, MetadataMetricType.NUMERIC_MEAN, MetadataMetricType.NUMERIC_STD,
    ]

    TEXT_METRICS = [
        MetadataMetricType.MIN_LENGTH, MetadataMetricType.MAX_LENGTH
    ]

    def __init__(
            self,
            data_source: "DataSource",
            dialect,
            policy: Policy,
            resource_name: str,
    ):
        self.resource_name = resource_name
        self.metadata_query_builder = MetadataQueryBuilder(dialect)
        super().__init__(
            # TODO: This is a hack to get the S3 extractor to work. We need to abstract this out
            S3Extractor(data_source.get_connection_string()),
            policy,
            dialect

        )

    def get_row_count(self):
        query = self.metadata_query_builder.row_count(self.resource_name)
        metadata = self.extractor.run(query)
        return metadata["rows"][0][metadata["columns"][0]]

    def extract_result(self, result, column_result):
        column_metrics = result["columns"]
        for column_metric in column_metrics:
            column, metric = column_metric.split("__")
            column = column.lower()
            metric = metric.lower()
            for row in result["rows"]:
                if isinstance(row[column_metric], decimal.Decimal):
                    column_result[column][metric] = round(float(row[column_metric]), 2)
                elif isinstance(row[column_metric], float):
                    column_result[column][metric] = round(row[column_metric], 2)
                else:
                    column_result[column][metric] = row[column_metric]
        return column_result

    def get_common_metadata(self, column_result: dict):
        # Get CARDINALITY AND NULL PERCENTAGE
        query = self.metadata_query_builder.get_metrics(
            list(column_result.keys()), self.resource_name, self.COMMON_METRICS)
        return self.extract_result(self.extractor.run(query), column_result)

    def get_numeric_metadata(self, column_result: dict, column_with_inferred_types: dict):
        numeric_columns = [col for col, col_type in column_result.items(
        ) if column_with_inferred_types[col] == "NUMERIC"]
        if len(numeric_columns) == 0:
            return column_result
        query = self.metadata_query_builder.get_metrics(
            numeric_columns, self.resource_name, self.NUMERIC_METRICS)
        return self.extract_result(self.extractor.run(query), column_result)

    def text_metadata(self, column_result: dict, column_with_inferred_types: dict):
        text_columns = [col for col, col_type in column_result.items(
        ) if column_with_inferred_types[col] == "TEXT"]
        if len(text_columns) == 0:
            return column_result
        query = self.metadata_query_builder.get_metrics(
            text_columns, self.resource_name, self.TEXT_METRICS)
        return self.extract_result(self.extractor.run(query), column_result)

    def get_metadata(self):
        discovery_data = self.get_discovery_data()
        column_with_inferred_types = {}
        column_result = {}
        resource_result = {}
        for row in discovery_data["rows"]:
            column_name = row["column_name"].lower()
            column_type = row["column_type"]
            column_result[column_name] = {}
            common_type = map_type_to_standard_types.get(column_type.upper(), "UNKNOWN")
            column_with_inferred_types[column_name] = infer_type(
                column_name, common_type)

        resource_result["row_count"] = self.get_row_count()
        column_result = self.get_common_metadata(column_result)
        column_result = self.get_numeric_metadata(
            column_result, column_with_inferred_types)
        column_result = self.text_metadata(column_result, column_with_inferred_types)

        return {
            "columns": column_result,
            "resource": resource_result,
        }

    def get_debugdata(self, debug_sql):
        debug_sql = self.get_debug_data(debug_sql)

        return debug_sql



from enum import Enum


class MetadataMetricType(Enum):
    """
    List of supported performance metrics
    """

    CARDINALITY = "_cardinality"
    NULL_PERCENTAGE = "_null_percentage"
    MAX_LENGTH = "_max_length"
    MIN_LENGTH = "_min_length"
    NUMERIC_MAX = "_numeric_max"
    NUMERIC_MIN = "_numeric_min"
    NUMERIC_MEAN = "_numeric_mean"
    NUMERIC_STD = "_numeric_std"



from enum import Enum
from typing import Dict, Optional, Type

import click

from altimate_models.base.source import DataSource
from altimate_models.base.ssh_source import SSHSource
from altimate_models.postgresql.source import PostgresSource
from altimate_models.s3.source import S3Source
from altimate_models.snowflake.source import SnowflakeSource
from altimate_models.tableau.source import TableauSource


class DataStoreType(Enum):
    SNOWFLAKE = "snowflake"
    POSTGRES = "postgres"
    S3 = "s3"
    TABLEAU = "tableau"


class DataStoreConfigFactory:
    CLASSES: dict[DataStoreType, Type[DataSource]] = {
        DataStoreType.SNOWFLAKE: SnowflakeSource,
        DataStoreType.POSTGRES: PostgresSource,
        DataStoreType.S3: S3Source,
        DataStoreType.TABLEAU: TableauSource,
    }

    @classmethod
    def prompt(cls, data_store_config: Dict) -> DataSource:
        data_store = cls.CLASSES[DataStoreType(data_store_config["type"])].prompt(
            data_store_config
        )
        has_jumphost = click.confirm("Do you have a jumphost?", default=False)
        if has_jumphost:
            jumphost = click.prompt("jump_hostname", type=str)
            jumpuser = click.prompt("jump_user", type=str)
            private_key_path = click.prompt("private_key_path", type=str)
            private_key_pass = click.prompt(
                "private_key_password", type=str, hide_input=True
            )
            jump_info = {
                "SSH_PKEY_PATH": private_key_path,
                "SSH_PKEY_PW": private_key_pass,
                "SSH_HOST": jumphost,
                "SSH_USER": jumpuser,
            }
            jump_data_store = SSHSource(jump_info=jump_info, original_ds=data_store)
            # test_data_store.test_connection()
            return jump_data_store

        else:
            return data_store

    @classmethod
    def create(cls, data_store_config: Dict) -> DataSource:
        data_store_config = cls.CLASSES[
            DataStoreType(data_store_config["type"])
        ].map_config_to_source(data_store_config)
        return cls.CLASSES[DataStoreType(data_store_config["type"])](
            **data_store_config
        )

    @classmethod
    def create_from_connection_string(
        cls,
        type: str,
        name: str,
        connection_string: str,
        key: str,
        jump_info: Optional[Dict] = None,
        key_pair: Optional[Dict] = None,
        auth_type: Optional[str] = None,
    ) -> DataSource:
        class_obj: Type[DataSource] = cls.CLASSES[DataStoreType(type)]
        configs = class_obj.get_config_from_url(connection_string)
        configs["name"] = name
        configs["type"] = type
        configs["key_pair"] = key_pair
        configs["auth_type"] = auth_type
        return_obj: DataSource = class_obj(**configs)
        return_obj.update_credentials(key)
        if jump_info:
            return_obj = SSHSource(jump_info=jump_info, original_ds=return_obj)
        return return_obj



from enum import Enum


class PerformanceMetricType(Enum):
    """
    List of supported performance metrics
    """

    MAX_EXECUTION_TIME = "_max_execution_time"
    MIN_BYTES_SPILLED_TO_LOCAL_STORAGE = "_min_bytes_spilled_to_local_storage"
    MAX_CREDITS_CONSUMED_BY_QUERY = "_max_credits_consumed_by_query"
    MAX_CREDITS_CONSUMED_BY_WAREHOUSE = "_max_credits_consumed_by_warehouse"



"""
This file has code that defines several functions that are used to fetch metadata information, such as the schema and profiler, of a data source.  The schema_mapper and profile_mapper functions are used to convert the data into a format that is expected by a contract definition. The fetch_schema and fetch_profile functions are responsible for loading the configuration and calling the appropriate functions to retrieve the metadata information. Additionally, the code defines several helper functions that are used to get metrics specific to certain column types (e.g. "int", "float", "string", "number") and others for semantic types (e.g. ID, date).
"""
from typing import Dict, List

from altimate_models import DataSource, ProfileConfig
from altimate_models.constants import COLUMN_TO_PERFORMANCE
from altimate_profiler import get_metrics_profile, get_schema_profile
from altimate_profiler.validator.mappers import schema_mapper


def gen_contract_for_performance(profile):
    output = {}
    for column, metric in profile.items():
        for name, value in metric.items():
            output[
                COLUMN_TO_PERFORMANCE[
                    (
                        column,
                        name,
                    )
                ]
            ] = value
    return output


def fetch_schema(
    data_source: DataSource,
    profile_config: ProfileConfig,
    use_account_usage: bool = False,
) -> List[Dict]:
    """
    Function which loads config and gets the schema for the source
    """
    original_schema = get_schema_profile(
        data_source=data_source, profile_config=profile_config, use_account_usage=use_account_usage
    )
    return schema_mapper(original_schema)


def fetch_profile(data_source: DataSource, profile_config: ProfileConfig) -> Dict:
    """
    Function which loads config and gets the Profile for the source
    """
    return {
        "policies": get_metrics_profile(
            data_source=data_source, profile_config=profile_config
        )
    }



import sys
from typing import Dict, Optional, Text

try:
    import click
    from sqlalchemy import create_engine
except ImportError:
    pass
from urllib.parse import quote_plus

from altimate_models.base.source import DataSource
from altimate_models.base.utils import load_private_key
from altimate_models.shared_models import (
    PrivateKeyType,
    Resource,
    ResourceType,
    TableResource,
)
from altimate_profiler.exceptions import AltimateDataStoreConnectionException
from altimate_profiler.utils import green_bold_string, red, sha256_encoding


class SnowflakeSource(DataSource):
    NAME = "snowflake"
    account: Optional[Text]
    role: Optional[Text]
    warehouse: Optional[Text]
    user: Optional[Text]
    password: Optional[Text]
    database: Optional[Text]
    sync_credentials: bool = False

    def get_connection_string(self, *args, **kwargs) -> Text:
        connection_string = "{NAME}://{user}:{password}@{account}/{database}".format(
            NAME=self.NAME,
            user=self.user,
            password=quote_plus(self.password),
            account=self.account,
            database=self.database,
            warehouse=self.warehouse,
            role=self.role,
        )
        # uncomment this when we add support for changing schema
        # CAUTION: schema is used internally by pydantic.
        # when we do add this, use a different name
        # if schema:
        #     connection_string += "/{schema}".format(schema=schema)
        connection_string += "?warehouse={warehouse}&role={role}".format(
            warehouse=self.warehouse,
            role=self.role,
        )
        return connection_string

    @classmethod
    def get_connection_information(cls, data_store_config: Dict):
        data_store_config[
            "connection_info"
        ] = f"{data_store_config['account']}/{data_store_config['database']}?role={data_store_config['role']}&warehouse={data_store_config['warehouse']}"
        data_store_config["connection_hash"] = sha256_encoding(
            data_store_config["connection_info"]
        )
        return data_store_config

    @classmethod
    def prompt(cls, data_store_config: Dict):
        if "click" not in sys.modules:
            raise ImportError(
                "click is not installed. Please install click to connect to S3."
            )
        data_store_config["account"] = click.prompt("account", type=str)
        data_store_config["role"] = click.prompt(
            "role", type=str, default="ACCOUNTADMIN"
        )
        data_store_config["warehouse"] = click.prompt("warehouse", type=str)
        data_store_config["user"] = click.prompt("user", type=str)
        data_store_config["password"] = click.prompt(
            "password", type=str, hide_input=True
        )
        data_store_config["database"] = click.prompt("database", type=str)
        data_store_config["sync_credentials"] = click.confirm(
            "sync credentials with altimate?", default=False
        )
        data_store_config = cls.get_connection_information(data_store_config)
        return SnowflakeSource(**data_store_config)

    def cred_prompt(self):
        if "click" not in sys.modules:
            raise ImportError(
                "click is not installed. Please install click to connect to S3."
            )
        self.user = click.prompt("user", type=str)
        self.password = click.prompt("password", type=str, hide_input=True)
        self.sync_credentials = click.confirm(
            "sync credentials with altimate?", default=False
        )
        conn_info = self.get_connection_information(self.__dict__)
        self.connection_info = conn_info["connection_info"]
        self.connection_hash = conn_info["connection_hash"]

    def get_databases(self):
        """
        Fetch all available databases from Snowflake using the INFORMATION_SCHEMA.DATABASES view.
        """
        try:
            with self.get_engine() as engine:
                with engine.connect() as connection:
                    # Use the configured warehouse
                    connection.execute(f"USE WAREHOUSE {self.warehouse}")
                    result = connection.execute(
                        """
                            SELECT
                                database_name
                            FROM
                            SNOWFLAKE.ACCOUNT_USAGE.DATABASES
                            WHERE deleted is null
                            AND database_name not in ('SNOWFLAKE', 'SNOWFLAKE_SAMPLE_DATA')
                            AND database_name not like 'USER$%'

                        """
                    )
                    databases = [dict(row) for row in result]
                    return databases
        except Exception as e:
            print("Failed to fetch databases from Snowflake.")
            print("Check your connection and credentials, then try again.")
            raise e

    def test_connection(self):
        if not "sqlalchemy" in sys.modules:
            raise ImportError(
                "sqlalchemy is not installed. Please install sqlalchemy to test connections"
            )
        try:
            with self.get_engine() as engine:
                engine.connect()
                print(green_bold_string("Successfully connected to Snowflake!"))

        except Exception as e:
            print(red(f"Connection to Snowflake failed. Error: {e}"))
            print("Check your credentials and try again.")
            raise AltimateDataStoreConnectionException(
                f"Connection test for user {self.user} failed for account {self.account}."
            )

    def drop_credentials(self):
        config_dict = self.dict()
        password = config_dict.pop("password")
        username = config_dict.pop("user")
        if self.sync_credentials:
            config_dict["key"] = f"altimate:{username}:{password}"
        return config_dict

    def get_resource(self):
        if self.database is None:
            raise ValueError("database must be set to get resource")
        return Resource(
            resource_type=ResourceType.TABLE,
            resource=TableResource(
                database=self.database,
            ),
        )

    def get_identifier(self):
        if "click" not in sys.modules:
            raise ImportError(
                "click is not installed. Please install click to connect to S3."
            )
        schema = click.prompt("schema", str)
        table = click.prompt("table", str)
        identifier = f"{self.NAME}://{self.database}/{schema}/{table}"
        fqn = f"database={self.database}/schema={schema}/table={table}"
        return identifier, fqn

    @classmethod
    def get_resource_config(cls, resource_config: Dict):
        filters = resource_config.get("filters")
        identifier = resource_config.get("identifier")
        if identifier is None:
            raise ValueError("identifier must be set to get resource")
        data_store_type, path = identifier.split("://")
        database, schema, table = path.split(".")
        resource = TableResource(database=database, schema=schema, table=table)
        return Resource(
            resource_type=ResourceType.TABLE, resource=resource, filters=filters
        )

    def get_identifier_and_fqn(self):
        if "click" not in sys.modules:
            raise ImportError(
                "click is not installed. Please install click to connect to S3."
            )
        schema = click.prompt("schema", type=str)
        table = click.prompt("table", type=str)
        identifier = f"{self.NAME}://{self.database}.{schema}.{table}"
        fqn = f"database={self.database}/schema={schema}/table={table}"
        return identifier, fqn

    @classmethod
    def map_config_to_source(cls, config):
        connection_info = config.get("connection_info")
        if not connection_info:
            return config
        snowflake_config = cls.get_config_from_url(connection_info)
        for k, v in snowflake_config.items():
            config[k] = v
        config.pop("connection_info")
        return config

    @classmethod
    def get_config_from_url(cls, connection_info):
        url, params = connection_info.split("?")
        account, database = url.split("/")
        config = {"account": account, "database": database}
        for param in params.split("&"):
            key, value = param.split("=")
            config[key] = value
        return config

    def update_credentials(self, credentials: str):
        _user, _pass = credentials.split(":")
        self.user = _user
        self.password = _pass

    def override_schema(self, schema):
        return schema

    def get_connect_args(self):
        connect_args = super().get_connect_args()
        if self.statement_timeout_in_ms is not None:
            connect_args["session_parameters"] = {
                "STATEMENT_TIMEOUT_IN_SECONDS": self.statement_timeout_in_ms // 1000
            }
        if self.auth_type == "key_pair":
            connect_args["private_key"] = load_private_key(
                self.key_pair["private_key"],
                self.key_pair.get("type", PrivateKeyType.PEM.value),
                self.key_pair.get("passphrase"),
            )
        return connect_args



from typing import TYPE_CHECKING

from altimate_models.base.profiler import SQLAlchemyProfiler
from altimate_models.postgresql.dialect import PostgreSQLSourceDialect
from altimate_models.postgresql.extractor import PostgreSQLExtractor
from altimate_models.shared_models import Policy, ResourceType

if TYPE_CHECKING:
    from altimate_models.postgresql.source import PostgresSource


class PostgreSQLProfiler(SQLAlchemyProfiler):
    def __init__(self, data_source: "PostgresSource", policy: Policy, use_account_usage: bool = False):
        if policy.resource0.resource_type != ResourceType.TABLE:
            raise Exception("PostgreSQLProfiler only supports table resources")

        # FIXME - There should be a check here to ensure that
        # database in source and in policy both match
        super().__init__(
            extractor=PostgreSQLExtractor(
                connection_str=data_source.get_connection_string(
                    # database=policy.resource0.resource.database,
                    # schema=policy.resource0.resource.db_schema,
                ),
                data_source=data_source,
            ),
            policy=policy,
            dialect=PostgreSQLSourceDialect(),
            use_account_usage=use_account_usage,
        )



import decimal
from uuid import uuid4

from altimate_profiler.transformer.base import Transformer


class MetricTransformer(Transformer):
    """
    Transforms the metrics collected by the MetricsCollectorJob into JSON Output
    """

    @classmethod
    def _original_schema(cls):
        return {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "type": "object",
            "properties": {
                "rows": {"type": "array", "items": {"type": "object"}, "minItems": 1},
                "columns": {"type": "array", "items": {"type": "string"}},
            },
            "required": ["rows", "columns"],
        }

    @classmethod
    def _normalized_schema(cls):
        """
        TODO: Value of the metric value should be of the correct type
        """
        return {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "id": {"type": "string"},
                    "table_name": {"type": "string"},
                    "schema": {"type": "string"},
                    "database": {"type": "string"},
                    "column_name": {"type": "string"},
                    "metric": {"type": "string"},
                    "value": {"type": "string"},
                    "time_window_start": {"type": "string"},
                    "time_window_end": {"type": "string"},
                },
                "required": ["id"],
            },
            "minItems": 1,
        }

    @classmethod
    def _transform(cls, inputs):
        """
        Transform the input into the normalized schema.
        """
        result = []
        for input in inputs:
            for row in input["rows"]:
                for column in input["columns"]:
                    result.append(
                        {
                            "is": float(row[column])
                            if type(row[column]) == decimal.Decimal
                            else row[column],
                        }
                    )
        return result

    @classmethod
    def _after_transform(cls, input):
        """
        Attach id to each row.
        """
        for metric in input:
            metric["id"] = uuid4().hex
        return {"terms": input}



from uuid import uuid4

from altimate_profiler.transformer.schema import SchemaTransformer


class DuckDBSchemaTransformer(SchemaTransformer):
    """
        Transforms the schema collected by the SchemaCollectorJob into JSON Output
        """

    @classmethod
    def _original_schema(cls):
        return {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "type": "object",
            "properties": {
                "rows": {"type": "array", "items": {"type": "object"}, "minItems": 1},
                "columns": {"type": "array", "items": {"type": "string"}},
            },
            "required": ["rows", "columns"],
        }

    @classmethod
    def _normalized_schema(cls):
        """
        We can maybe fetch the tags etc here.
        """
        return {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "table_name": {"type": "string"},
                    "schema": {"type": "string"},
                    "database": {"type": "string"},
                    "column_name": {"type": "string"},
                    "column_type": {"type": "string"},
                },
                "required": ["id"],
            },
            "minItems": 1,
        }

    @classmethod
    def _transform(cls, input):
        results = []
        # Renaming the values
        # TODO: This is a hack, we should do this in the discovery query
        # Creating ticket for it.
        for row in input["rows"]:
            results.append({
                "COL_NAME": row.get("column_name"),
                "COL_TYPE": row.get("column_type"),
                "FILE_PATH": row.get("file_path"),
                "FILE_FORMAT": row.get("file_format"),
                "BUCKET": row.get("bucket"),
            })
        return results

    @classmethod
    def _after_transform(cls, input):
        """
        Attach id to each row.
        """
        for column in input:
            column["id"] = uuid4().hex
        return input



from typing import List, Text

from altimate_models.shared_models import Policy
from altimate_profiler.metrics.column_metric_type import ColumnMetricType
from altimate_profiler.builders.base_builder import Base


class S3ColumnMetricsQueryBuilder(Base):

    def __init__(
            self,
            dialect,
            policy: Policy
    ):
        self.dialect = dialect
        self.policy = policy

    def _gen_sql(self, metric, params, alias):
        attr = getattr(self.dialect, metric._value_)

        if not attr:
            raise Exception(
                "Unreachable: Metric type is defined that does not resolve to a definition."
            )
        select_unformatted = attr()
        select_no_alias = select_unformatted.format(**params)
        select = "{} AS {}".format(select_no_alias, alias)
        return select

    def compile(self) -> List[Text]:
        metrics_query = []

        for t_idx, term in enumerate(self.policy.terms):
            metrics_query.append(
                self._gen_sql(ColumnMetricType[term.metric.upper()], term.params, f"term_{t_idx}")
            )

        # aggregate different level of filters
        all_filters = ' AND '.join(x for x in [self.policy.resource0.filters, self.policy.filters] if x)

        select_sql = ",\n".join(metrics_query)
        query = self.dialect._base_query_sample(
            select_sql=select_sql,
            bucket=self.policy.resource0.resource.bucket,
            path=self.policy.resource0.resource.path,
            options=self.policy.resource0.resource.options,
            filters=all_filters
        )

        return [query]




from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from altimate_models.base.source import DataSource
from altimate_models.s3.csv_source_dialect import CsvSourceDialect
from altimate_models.s3.json_source_dialect import JsonSourceDialect
from altimate_models.s3.parquet_source_dialect import ParquetSourceDialect
from altimate_models.duckdb.profiler import DuckDBProfiler
from altimate_models.shared_models import Policy
from altimate_models.s3.extractor import S3Extractor


class S3Profiler(DuckDBProfiler):
    dialects = {
        "csv": CsvSourceDialect,
        "parquet": ParquetSourceDialect,
        "json": JsonSourceDialect,
    }

    def __init__(self, data_source: "DataSource", policy: Policy, use_account_usage: bool = False):
        super().__init__(
            extractor=S3Extractor(
                data_source.get_connection_string()),
            policy=policy,
            dialect=self.dialects[policy.resource0.resource.format](),
            use_account_usage=use_account_usage,
        )





from typing import List, Text

from altimate_models.shared_models import Policy
from altimate_profiler.builders.base_builder import Base
from altimate_profiler.metrics.performance_metric_type import PerformanceMetricType


class PerformanceQueryBuilder(Base):
    """
    Query builder for performance metrics
    TODO: As of right now, multiple queries are executed. We should be able to do this in one query
    """

    def __init__(self, dialect, policy: Policy):
        self.dialect = dialect
        self.policy = policy

    def _gen_sql(self, metric_name, params) -> Text:
        attr = getattr(self.dialect, PerformanceMetricType[metric_name.upper()].value)
        return attr(params)

    def compile(self) -> List[Text]:
        metrics_queries = []

        for term in self.policy.terms:
            metrics_queries.append(self._gen_sql(term.metric, term.params))

        return metrics_queries



from typing import Text

map_type_to_standard_types = {
    "BOOLEAN": "BOOLEAN",
    "ENUM": "ENUM",
    "NUMERIC": "NUMERIC",
    "VARCHAR": "TEXT",
    "DOUBLE": "NUMERIC",
    "TEXT": "TEXT",
    "DATE": "DATE",
    "TIMESTAMP": "TIMESTAMP",
    "TIME": "TIMESTAMP",
    "INTERVAL": "INTERVAL",
    "BLOB": "TEXT",
    "BITSTRING": "TEXT",
    "FLOAT": "NUMERIC",
    "NUMBER": "NUMERIC",
    "LIST": "LIST",
    "STRUCT": "STRUCT",
    "MAP": "MAP",
    "BIGINT": "INTEGER",
    "CHARACTER VARYING": "TEXT",
    "DOUBLE PRECISION": "NUMERIC",
    "INTEGER": "INTEGER",
}


def infer_type(column_name: Text, column_type: Text):
    # Check if column is an id
    if column_type in ["NUMERIC", "TEXT"]:
        if "id" in column_name.lower():
            return "ID"

    # Check if column is a timestamp
    if column_type in ["TIMESTAMP", "DATE"]:
        return "DATE"
    elif column_type == "TEXT":
        if "date" in column_name.lower():
            return "DATE"

    return column_type



from dataclasses import dataclass
from typing import List, Text

from altimate_models.base.dialect import SQLAlchemySourceDialect
from altimate_models.base.extractor import SQLAlchemyExtractor
from altimate_models.shared_models import Policy
from altimate_models.snowflake.dialect import SnowflakeSourceDialect
from altimate_profiler.builders.custom_query_builder import CustomQueryBuilder
from altimate_profiler.builders.performance_query_builder import PerformanceQueryBuilder
from altimate_profiler.builders.schema_query_builder import SchemaQueryBuilder


@dataclass
class SQLAlchemyProfiler:
    extractor: SQLAlchemyExtractor
    policy: Policy
    dialect: SQLAlchemySourceDialect
    use_account_usage: bool = False

    def _columns_schema(self, _) -> Text:
        return self.dialect.schema_columns_query(
            database_name=self.policy.resource0.resource.database,
            schema_name=self.policy.resource0.resource.db_schema,
            table_name=self.policy.resource0.resource.table,
            filters=self.policy.resource0.resource.filters,
        )

    def _tables_schema(self, _) -> Text:
        return self.dialect.schema_tables_query(
            database_name=self.policy.resource0.resource.database,
            schema_name=self.policy.resource0.resource.db_schema,
            table_name=self.policy.resource0.resource.table,
            filters=self.policy.resource0.resource.filters,
        )

    def _access_logs(self, _) -> Text:
        return self.dialect.access_logs_query()

    def _copy_and_load_logs(self, _) -> Text:
        return self.dialect.copy_and_load_logs_query()

    def _discovery_query(self) -> Text:
        sql = self.dialect.schema_columns_query(
            database_name=self.policy.resource0.resource.database,
            schema_name=self.policy.resource0.resource.db_schema,
            table_name=self.policy.resource0.resource.table,
            use_account_usage=self.use_account_usage,
        )
        return sql

    def _metrics_query(self, discovery_data) -> List[Text]:
        queries = self.dialect.column_metrics_query(discovery_data, self.policy)

        return queries

    def _custom_query(self) -> List[Text]:
        custom_metric_builder = CustomQueryBuilder(self.policy)
        return custom_metric_builder.compile()

    def get_discovery_data(self):
        return self.extractor.run(self._discovery_query())

    def get_debug_data(self, debug_sql):
        return self.extractor.run(debug_sql)

    def _table_query(self):
        return self.dialect.table_metrics_query(self.policy)

    def _performance_query(self):
        if type(self.dialect) != SnowflakeSourceDialect:
            raise Exception("Performance check is only supported for Snowflake")
        performance_query_builder = PerformanceQueryBuilder(self.dialect, self.policy)
        return performance_query_builder.compile()

    def _schema_policy(self):
        discovery_data = self.get_discovery_data()
        result = SchemaQueryBuilder(self.policy, discovery_data).compile()
        return result

    def get_metrics_data(self):
        # TODO: Create Policy Type Abstract Class
        discovery_data = self.get_discovery_data()
        if self.policy.policy_type == "custom_check":
            metrics_sql = self._custom_query()
        elif self.policy.policy_type == "column":
            metrics_sql = self._metrics_query(discovery_data)
        elif self.policy.policy_type == "table":
            metrics_sql = self._table_query()
        elif self.policy.policy_type == "performance":
            metrics_sql = self._performance_query()
        elif self.policy.policy_type == "schema":
            # TODO: Make this better
            result = self._schema_policy()
            return result
        else:
            raise Exception("Unknown policy type")

        metrics = []
        for sql in metrics_sql:
            metric = self.extractor.run(sql)
            metrics.append(metric)

        return metrics

    def get_table_metadata(self):
        return self.get_metadata(self._tables_schema)

    def get_metadata(self, schema):
        table_metadata = self.extractor.run(self._tables_schema(schema))
        for column_name, column_type in table_metadata.items():
            column_metadata = self.extractor.run(self._columns_schema(column_name))
            table_metadata[column_name] = {
                "type": column_type,
                "metadata": column_metadata,
            }



from io import BytesIO

try:
    from cryptography.hazmat.primitives import serialization
    from cryptography.hazmat.backends import default_backend
except ImportError:
    pass


def get_load_function(key_type: str):
    if key_type == 'pem':
        return serialization.load_pem_private_key
    elif key_type == 'der':
        return serialization.load_der_private_key
    else:
        raise ValueError(f'Unsupported key type: {key_type}')

def load_private_key(
    private_key_string: str,
    private_key_type: str = 'pem',
    passphrase: str = None,
):
    load_func = get_load_function(private_key_type)
    key_bytes = private_key_string.encode('utf-8')

    if passphrase is not None:
        passphrase = passphrase.encode('utf-8') if isinstance(passphrase, str) else passphrase

    # Use BytesIO to wrap the bytes as a file-like object
    key_file_like = BytesIO(key_bytes)
    p_key = load_func(
        key_file_like.read(),
        password=passphrase,
        backend=default_backend()
    )
    pkb = p_key.private_bytes(
        encoding=serialization.Encoding.DER,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption()
    )
    return pkb



DEFAULT_CLIENT_DB_CONNECTION_TIMEOUT = 5

PERFORMANCE_TO_COLUMN = {
    "AVG_QUERY_TIME_MS": ("EXECUTION_TIME", "NUMERIC_MEAN"),
    "MAX_QUERY_TIME_MS": ("EXECUTION_TIME", "NUMERIC_MAX"),
    "AVG_CREDITS_USED_CLOUD_SERVICES": ("CREDITS_USED_CLOUD_SERVICES", "NUMERIC_MEAN"),
    "PERCENTAGE_OF_QUERIES_WITH_ERRORS": ("EXECUTION_STATIS", "PERCENTAGE_FAILURES"),
    "PERCENTAGE_OF_SELECTS_WITH_WAREHOUSE_XSMALL": (
        "QUERY_TYPE",
        "PERCENTAGE_WAREHOUSE_IS_XSMALL",
    ),
    "AVG_NUMBER_OF_ROWS_INSERTED_FOR_INSERT_QUERY": (
        "ROWS_INSERTED",
        "AVG_ROWS_INSERTED_OR_UPDATED",
    ),
    "AVG_BYTES_SPILLED_TO_LOCAL_STORAGE": (
        "BYTES_SPILLED_TO_LOCAL_STORAGE",
        "NUMERIC_MEAN",
    ),
    "AVG_BYTES_SPILLED_TO_REMOTE_STORAGE": (
        "BYTES_SPILLED_TO_REMOTE_STORAGE",
        "NUMERIC_MEAN",
    ),
    "PERCENTAGE_OF_EXPLORATORY_QUERIES": ("QUERY_TYPE", "PERCENTAGE_SELECTS"),
    "AVG_TOTAL_ELAPSED_TIME": ("TOTAL_ELAPSED_TIME", "AVERAGE_NON_NEGATIVE"),
    "MIN_TOTAL_ELAPSED_TIME": ("TOTAL_ELAPSED_TIME", "MIN_NON_NEGATIVE"),
    "MAX_TOTAL_ELAPSED_TIME": ("TOTAL_ELAPSED_TIME", "MAX_NON_NEGATIVE"),
    "AVG_SELECTION_QUERY_TIME_MS": ("TOTAL_ELAPSED_TIME", "AVG_SELECT_QUERY_TIME"),
}

COLUMN_TO_PERFORMANCE = {v: k for k, v in PERFORMANCE_TO_COLUMN.items()}


TEST_SQL_QUERY = "SELECT 1"



class AltimateProfilerException(Exception):
    """
    Raised when the Altimate Profiler encounters an error.
    """

    pass


class AltimateDataStoreNotSupported(Exception):
    """
    Raised when the data store is not supported
    """

    pass


class AltimateDataStoreConnectionException(Exception):
    """
    Raised when the data store is not supported
    """

    pass


class AltimateInvalidInputException(Exception):
    pass


INVALID_UNICODE_IN_CSV = "Invalid unicode (byte sequence mismatch) detected in CSV file"


class AltimateInvalidCharactersInCSVException(Exception):
    pass



from typing import List, Text

from altimate_models.base.dialect import SQLAlchemySourceDialect


class SnowflakeSourceDialect(SQLAlchemySourceDialect):
    @classmethod
    def _max_execution_time(cls, params) -> Text:
        return (
            "SELECT COALESCE(MAX(EXECUTION_TIME/1000), 0) FROM snowflake.account_usage.query_history"
            " WHERE END_TIME >= DATEADD('hour', -{hours}, CURRENT_TIMESTAMP())".format(
                hours=params.get("hours", 1)
            )
        )

    @classmethod
    def _min_bytes_spilled_to_local_storage(cls, params) -> Text:
        return (
            "SELECT  COALESCE(MIN(BYTES_SPILLED_TO_LOCAL_STORAGE), 0) FROM snowflake.account_usage.query_history"
            " WHERE END_TIME >= DATEADD('hour', -{hours}, CURRENT_TIMESTAMP())".format(
                hours=params.get("hours", 1)
            )
        )

    @classmethod
    def _cardinality(cls):
        return "COUNT(DISTINCT md5({column}))"

    @classmethod
    def _null_percentage(cls):
        return "SUM(IFF({column} IS NULL, 1 , 0))/COALESCE(COUNT(*), 1) *100"

    @classmethod
    def _max_credits_consumed_by_warehouse(cls, params) -> Text:
        return (
            "SELECT COALESCE(MAX(CREDITS_USED), 0) FROM snowflake.account_usage.warehouse_metering_history"
            " WHERE END_TIME >= DATEADD('hour', -{hours}, CURRENT_TIMESTAMP())".format(
                hours=params.get("hours", 1)
            )
        )

    @classmethod
    def _duplicate_count(cls, table, params):
        return "SELECT COALESCE(COUNT(c), 0) AS duplicates FROM (SELECT HASH(ARRAY_TO_STRING(ARRAY_CONSTRUCT({columns}), ',')) AS hash_value, COUNT(*) AS c FROM {table} GROUP BY hash_value HAVING COUNT(*) > 1) AS t".format(
            columns=",".join(params["columns"]), table=table
        )

    @classmethod
    def _null_count(cls):
        return "SUM(IFF({column} IS NULL, 1 , 0))"

    @classmethod
    def _regex_match_percentage(cls):
        return "SUM(IFF({column}::TEXT REGEXP '{pattern}', 1 , 0)) / COALESCE(COUNT(*), 1) * 100"

    @classmethod
    def _like_percentage(cls):
        return "SUM(IFF({column}:: TEXT  LIKE '{pattern}', 1 , 0)) / COALESCE(COUNT(*), 1) * 100"

    @classmethod
    def _null_count_table(cls, table, params):
        condition = " AND ".join([f"{column} IS NULL" for column in params["columns"]])
        return f"SELECT SUM(IFF({condition}, 1, 0))  FROM {table}".format(
            table=table, condition=condition
        )

    @classmethod
    def _allowed_values(cls):
        return "SUM(IFF({column} NOT IN {values}, 1 , 0))"

    @classmethod
    def _max_credits_consumed_by_query(cls, params) -> Text:
        sql = """
            WITH
                warehouse_sizes AS (
                    SELECT 'X-Small' AS warehouse_size, 1 AS credits_per_hour UNION ALL
                    SELECT 'Small' AS warehouse_size, 2 AS credits_per_hour UNION ALL
                    SELECT 'Medium'  AS warehouse_size, 4 AS credits_per_hour UNION ALL
                    SELECT 'Large' AS warehouse_size, 8 AS credits_per_hour UNION ALL
                    SELECT 'X-Large' AS warehouse_size, 16 AS credits_per_hour UNION ALL
                    SELECT '2X-Large' AS warehouse_size, 32 AS credits_per_hour UNION ALL
                    SELECT '3X-Large' AS warehouse_size, 64 AS credits_per_hour UNION ALL
                    SELECT '4X-Large' AS warehouse_size, 128 AS credits_per_hour
                )
                SELECT
                    COALESCE(MAX(qh.execution_time/(1000*60*60)*wh.credits_per_hour) , 0) AS query_cost
                FROM snowflake.account_usage.query_history AS qh
                INNER JOIN warehouse_sizes AS wh
                    ON qh.warehouse_size=wh.warehouse_size
                WHERE
                    END_TIME >= DATEADD('hour', -{hours}, CURRENT_TIMESTAMP())
        
        """
        return sql.format(hours=params.get("hours", 1))

    @classmethod
    def _numeric_mode(cls):
        return "MODE({column})"

    @classmethod
    def _numeric_median(cls):
        return "MEDIAN({column})"

    @classmethod
    def _percentage_selects(cls):
        return "SUM(IFF(TO_VARCHAR({column}) = 'SELECT', 1, 0)) * 100 / COALESCE(COUNT(*), 1)"

    @classmethod
    def _percentile(cls):
        return "PERCENTILE_CONT({value}) WITHIN GROUP (ORDER BY {column})"

    @classmethod
    def _conditional_count(cls):
        return "SUM(IFF({condition}, 1, 0))"

    @classmethod
    def _conditional_percentage(cls):
        return "SUM(IFF({condition}, 1, 0)) * 100 / COALESCE(COUNT(*), 1)"

    @classmethod
    def _conditional_max(cls):
        return "MAX(IFF({condition}, {column}, NULL))"

    @classmethod
    def _conditional_min(cls):
        return "MIN(IFF({condition}, {column}, NULL))"

    @classmethod
    def _conditional_mean(cls):
        return "AVG(IFF({condition}, {column}, NULL))"

    @classmethod
    def _percentage_inserts(cls):
        return "SUM(IFF(TO_VARCHAR({column}) = 'INSERT', 1, 0)) * 100 / COALESCE(COUNT(*), 1)"

    @classmethod
    def _percentage_warehouse_size_x_small(cls):
        return "SUM(IFF(TO_VARCHAR({column}) = 'X-Small', 1, 0)) * 100 / COALESCE(COUNT(*), 1)"

    @classmethod
    def _percentage_failures(cls):
        return "SUM(IFF(TO_VARCHAR({column}) LIKE '%FAIL%', 1, 0)) * 100 / COALESCE(COUNT(*), 1)"

    @classmethod
    def _percentage_warehouse_is_xsmall(cls):
        return (
            "SUM(IFF( (TO_VARCHAR({column}) = 'SELECT') AND (WAREHOUSE_SIZE = 'X-Small'), 1, 0)) * 100"
            "/ SUM(IFF( (TO_VARCHAR({column}) = 'SELECT') AND (WAREHOUSE_SIZE IS NOT NULL), 1, 0))"
        )

    @classmethod
    def _avg_rows_inserted_or_updated(cls):
        return (
            "SUM(IFF( QUERY_TYPE = 'INSERT', "
            "ROWS_UPDATED + ROWS_INSERTED, 0)) / SUM(IFF( QUERY_TYPE = 'INSERT', 1, 0)) "
        )

    @classmethod
    def _average_non_negative(cls):
        """
        The execution time for the query_history column is negative for
        queries to the query_history table.
        """
        return "AVG(IFF({column} >= 0 , {column}, 0))"

    @classmethod
    def _avg_select_query_time(cls):
        return "AVG(IFF( QUERY_TYPE = 'SELECT', TOTAL_ELAPSED_TIME, 0))"

    @classmethod
    def _max_non_negative(cls):
        return "MAX(IFF({column} >= 0 , {column}, 0))"

    @classmethod
    def _min_non_negative(cls):
        return "MIN(IFF({column} >= 0 , {column}, 0))"

    @classmethod
    def _text_int_rate(cls):
        return (
            "SUM(IFF(REGEXP_COUNT(TO_VARCHAR({column}), '^([-+]?[0-9]+)$', 1, 'i') != 0, 1,"
            " 0)) / COALESCE(COUNT(*), 1)"
        )

    @classmethod
    def _text_number_rate(cls):
        return (
            "SUM(IFF(REGEXP_COUNT(TO_VARCHAR({column}),"
            " '^([-+]?[0-9]*[.]?[0-9]+([eE][-+]?[0-9]+)?)$', 1, 'i') != 0, 1, 0)) /"
            " COALESCE(COUNT(*), 1)"
        )

    @classmethod
    def _text_uuid_rate(cls):
        return (
            "SUM(IFF(REGEXP_COUNT(TO_VARCHAR({column}),"
            " '^([0-9a-fA-F]{{8}}-[0-9a-fA-F]{{4}}-[0-9a-fA-F]{{4}}-[0-9a-fA-F]{{4}}-[0-9a-fA-F]{{12}})$',"
            " 1, 'i') != 0, 1, 0)) / COALESCE(COUNT(*), 1)"
        )

    @classmethod
    def _text_all_spaces_rate(cls):
        return (
            "SUM(IFF(REGEXP_COUNT(TO_VARCHAR({column}), '^(\\\\s+)$', 1, 'i') != 0, 1, 0)) /"
            " COALESCE(COUNT(*), 1)"
        )

    @classmethod
    def _text_null_keyword_rate(cls):
        return (
            "SUM(IFF(UPPER({column}) IN ('NULL', 'NONE', 'NIL', 'NOTHING'), 1, 0)) /"
            " COALESCE(COUNT(*), 1)"
        )

    @classmethod
    def _zero_rate(cls):
        return (
            "SUM(IFF(UPPER({column}) IN ('NULL', 'NONE', 'NIL', 'NOTHING'), 1, 0)) /"
            " COALESCE(COUNT(*), 1)"
        )

    @classmethod
    def _negative_rate(cls):
        return "SUM(IFF({column} < 0, 1, 0)) / COALESCE(COUNT(*), 1)"

    @classmethod
    def _dates_in_future_percentage(cls):
        return "SUM(IFF({column} > CURRENT_TIMESTAMP(), 1, 0)) / COALESCE(COUNT(*), 1) * 100"

    @classmethod
    def _completeness(cls):
        return "COUNT({column}) / CAST(COUNT(*) AS NUMERIC)"

    @classmethod
    def _invalid_email_rate(cls):
        return (
            "SUM(IFF(REGEXP_COUNT(TO_VARCHAR({column}),"
            " '^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+.[a-zA-Z0-9\\\-.]+$',"
            " 1, 'i') != 0, 0, 1)) / COALESCE(COUNT(*), 1)"
        )

    @classmethod
    def _freshness(cls, table, params):
        column = params.get("column")
        time_unit = params.get("timeUnit")
        return "select min(datediff({time_unit}, {column}, current_timestamp())) from {table}".format(
            table=table, column=column, time_unit=time_unit
        )

    @classmethod
    def schema_tables_query(
        cls,
        database_name: Text,
        schema_name: Text,
        table_name: Text = None,
        filters: Text = None,
        schemas_to_exclude: List[Text] = ["information_schema"],
    ) -> Text:
        excluded_schemas_string = ", ".join(
            ["'{}'".format(schema) for schema in schemas_to_exclude]
        )
        return """
            SELECT 
              TABLE_CATALOG, 
              TABLE_SCHEMA,
              TABLE_NAME, 
              TABLE_OWNER, 
              TABLE_TYPE, 
              IS_TRANSIENT, 
              RETENTION_TIME, 
              AUTO_CLUSTERING_ON, 
              COMMENT 
            FROM {database_name}.information_schema.tables 
            WHERE 
              table_schema NOT IN ({excluded_schemas_string}) 
              AND TABLE_TYPE NOT IN ('VIEW', 'EXTERNAL TABLE') 
              AND LOWER( TABLE_SCHEMA ) = LOWER('{schema_name}')
            ORDER BY TABLE_CATALOG, TABLE_SCHEMA, TABLE_NAME;
        """.format(
            database_name=database_name,
            schema_name=schema_name,
            excluded_schemas_string=excluded_schemas_string,
        )

    @classmethod
    def schema_columns_query(
        cls,
        database_name: Text,
        schema_name: Text,
        table_name: Text = None,
        filters: Text = None,
        schemas_to_exclude: List[Text] = ["information_schema"],
        use_account_usage: bool = False,
    ) -> Text:
        
        if use_account_usage:
            # Build schema filter for subqueries
            excluded_schemas_string = ", ".join(
                ["'{}'".format(schema.upper()) for schema in schemas_to_exclude]
            )
            
            if schema_name:
                schema_filter = f"AND LOWER(table_schema) = LOWER('{schema_name}')"
            else:
                schema_filter = f"AND UPPER(table_schema) NOT IN ({excluded_schemas_string})"
            
            sql = """
                SELECT
                    lower(c.table_name) AS NAME,
                    lower(c.column_name) AS COL_NAME,
                    lower(c.data_type) AS COL_TYPE,
                    c.comment AS COL_DESCRIPTION,
                    lower(c.ordinal_position) AS COL_SORT_ORDER,
                    lower(c.table_catalog) AS DATABASE,
                    lower(c.table_schema) AS SCHEMA,
                    t.comment AS DESCRIPTION,
                    decode(lower(t.table_type), 'view', 'true', 'false') AS IS_VIEW
                FROM
                    (SELECT * FROM SNOWFLAKE.ACCOUNT_USAGE.COLUMNS 
                    WHERE LOWER(table_catalog) = LOWER('{database_name}') 
                    AND deleted IS NULL
                    {schema_filter}) AS c
                LEFT JOIN
                    (SELECT * FROM SNOWFLAKE.ACCOUNT_USAGE.TABLES 
                    WHERE LOWER(table_catalog) = LOWER('{database_name}') 
                    AND deleted IS NULL
                    {schema_filter}) t
                        ON c.TABLE_NAME = t.TABLE_NAME
                        AND c.TABLE_SCHEMA = t.TABLE_SCHEMA
                        AND c.TABLE_CATALOG = t.TABLE_CATALOG
            """.format(
                database_name=database_name, 
                schema_filter=schema_filter
            )

            # Add additional filters after the main query for account usage
            where_conditions = []
            
            if table_name:
                where_conditions.append(f"LOWER(NAME) = LOWER('{table_name}')")

            if filters:
                where_conditions.append(filters)

            if where_conditions:
                sql += " WHERE " + " AND ".join(where_conditions)
                
        else:
            # Original INFORMATION_SCHEMA approach
            sql = """
                SELECT
                    lower(c.table_name) AS NAME,
                    lower(c.column_name) AS COL_NAME,
                    lower(c.data_type) AS COL_TYPE,
                    c.comment AS COL_DESCRIPTION,
                    lower(c.ordinal_position) AS COL_SORT_ORDER,
                    lower(c.table_catalog) AS DATABASE,
                    lower(c.table_schema) AS SCHEMA,
                    t.comment AS DESCRIPTION,
                    decode(lower(t.table_type), 'view', 'true', 'false') AS IS_VIEW
                FROM
                    {database_name}.INFORMATION_SCHEMA.COLUMNS AS c
                LEFT JOIN
                    {database_name}.INFORMATION_SCHEMA.TABLES t
                        ON c.TABLE_NAME = t.TABLE_NAME
                        AND c.TABLE_SCHEMA = t.TABLE_SCHEMA
                WHERE LOWER( c.table_catalog ) = LOWER('{database_name}')
            """.format(
                database_name=database_name, schema_name=schema_name
            )

            excluded_schemas_string = ", ".join(
                ["'{}'".format(schema) for schema in schemas_to_exclude]
            )
            
            if schema_name:
                sql += f" AND LOWER( c.table_schema ) = LOWER('{schema_name}')"
            else:
                sql += f" AND LOWER( c.table_schema ) NOT IN ({excluded_schemas_string})"

            if table_name:
                sql += f" AND LOWER(c.table_name) = LOWER('{table_name}')"

            if filters:
                sql += f" AND {filters}"

        return sql

    @classmethod
    def access_logs_query(cls):
        return """
            SELECT 
                "QUERY_TEXT", 
                "DATABASE_NAME", 
                "SCHEMA_NAME", 
                "QUERY_TYPE", 
                "USER_NAME", 
                "ROLE_NAME", 
                "EXECUTION_STATUS", 
                "START_TIME", 
                "END_TIME", 
                "TOTAL_ELAPSED_TIME", 
                "BYTES_SCANNED", 
                "ROWS_PRODUCED", 
                "SESSION_ID", 
                "QUERY_ID", 
                "QUERY_TAG", 
                "WAREHOUSE_NAME", 
                "ROWS_INSERTED", 
                "ROWS_UPDATED", 
                "ROWS_DELETED", 
                "ROWS_UNLOADED" 
            FROM snowflake.account_usage.query_history 
            WHERE 
                start_time BETWEEN to_timestamp_ltz('2021-01-01 00:00:00.000000+00:00') AND to_timestamp_ltz('2021-01-01 01:00:00.000000+00:00') 
                AND QUERY_TYPE NOT IN ('DESCRIBE', 'SHOW') 
                AND (DATABASE_NAME IS NULL OR DATABASE_NAME NOT IN ('UTIL_DB', 'SNOWFLAKE')) 
                AND ERROR_CODE is NULL 
            ORDER BY start_time DESC;
        """

    @classmethod
    def copy_and_load_logs_query(cls):
        return """
            SELECT 
                "FILE_NAME", 
                "STAGE_LOCATION", 
                "LAST_LOAD_TIME", 
                "ROW_COUNT", 
                "FILE_SIZE", 
                "ERROR_COUNT", 
                "STATUS", 
                "TABLE_CATALOG_NAME", 
                "TABLE_SCHEMA_NAME", 
                "TABLE_NAME", 
                "PIPE_CATALOG_NAME", 
                "PIPE_SCHEMA_NAME", 
                "PIPE_NAME", 
                "PIPE_RECEIVED_TIME" 
            FROM snowflake.account_usage.copy_history 
            WHERE 
                LAST_LOAD_TIME between to_timestamp_ltz('2021-01-01 00:00:00.000000+00:00') AND to_timestamp_ltz('2021-01-01 01:00:00.000000+00:00')
                AND STATUS != 'load failed' 
            ORDER BY LAST_LOAD_TIME DESC;
        """
